open Program
open Interval

exception Error of string

type loc = id
type value = Value of int | Addr of loc

(* 231010 = 130 * 1777 *)
let d1 = 1777
let d2 = 130
let r1 = 1129 mod d1
let r2 = 1129 mod d2

let rem n d =
  let r = n mod d in
  if r < 0 then r + d else r

(* for P(Z/dZ) *)
module RemSet = Set.Make (struct
  type t = int

  let compare = compare
end)

module LocSet = Set.Make (struct
  type t = loc

  let compare = compare
end)

module LocMap = Map.Make (struct
  type t = loc

  let compare = compare
end)

module LblMap = Map.Make (struct
  type t = int

  let compare = compare
end)

type memory = value LocMap.t

module MemSet = Set.Make (struct
  type t = memory

  let compare = LocMap.compare (compare : value -> value -> int)
end)

type collecting_semantics = MemSet.t LblMap.t

let rec make_top_rem n acc =
  if n >= 0 then make_top_rem (n - 1) (RemSet.add n acc) else acc

let top_rem d = make_top_rem (d - 1) RemSet.empty
let join_rem r1 r2 = RemSet.union r1 r2
let widening_threshold = 12

let widen_rem d r1 r2 =
  let joined = join_rem r1 r2 in
  if RemSet.cardinal joined > widening_threshold then top_rem d else joined

let plus_rem d r1 r2 =
  let add_one n rems =
    RemSet.fold (fun m acc -> RemSet.add (rem (n + m) d) acc) rems RemSet.empty
  in
  RemSet.fold (fun n acc -> RemSet.union (add_one n r1) acc) r2 RemSet.empty

let mult_rem d r1 r2 =
  let mult_one n rems =
    RemSet.fold (fun m acc -> RemSet.add (rem (n * m) d) acc) rems RemSet.empty
  in
  RemSet.fold (fun n acc -> RemSet.union (mult_one n r1) acc) r2 RemSet.empty

let neg_rem d r =
  RemSet.fold (fun n acc -> RemSet.add (rem (-n) d) acc) r RemSet.empty

let rem_true = RemSet.singleton 1
let rem_false = RemSet.singleton 0
let rem_bool_top = RemSet.add 0 (RemSet.singleton 1)

let less_rem r1 r2 =
  if RemSet.is_empty r1 || RemSet.is_empty r2 then RemSet.empty
  else rem_bool_top

let maybe_f_rem r = RemSet.mem 0 r
let maybe_t_rem r = not (RemSet.is_empty r)

let not_rem r =
  match (maybe_f_rem r, maybe_t_rem r) with
  | false, false -> RemSet.empty (* bottom *)
  | false, true -> rem_false
  | true, false -> rem_true
  | true, true -> rem_bool_top

let and_rem r1 r2 =
  match (RemSet.is_empty r1, RemSet.is_empty r2) with
  | true, _ | _, true -> RemSet.empty
  | _ -> (
    let t1 = maybe_t_rem r1 in
    let t2 = maybe_t_rem r2 in
    let f1 = maybe_f_rem r1 in
    let f2 = maybe_f_rem r2 in
    match (t1 && t2, f1 || f2) with
    | true, true -> rem_bool_top
    | true, false -> rem_true
    | _ -> rem_false)

let or_rem r1 r2 =
  match (RemSet.is_empty r1, RemSet.is_empty r2) with
  | true, _ | _, true -> RemSet.empty
  | _ -> (
    let t1 = maybe_t_rem r1 in
    let t2 = maybe_t_rem r2 in
    let f1 = maybe_f_rem r1 in
    let f2 = maybe_f_rem r2 in
    match (t1 || t2, f1 && f2) with
    | true, true -> rem_bool_top
    | true, false -> rem_true
    | _ -> rem_false)

type abs_int = itv * RemSet.t * RemSet.t

let top_int = (top_itv, top_rem d1, top_rem d2)

let join_int (i1, r1, r1') (i2, r2, r2') =
  (join_itv i1 i2, join_rem r1 r2, join_rem r1' r2')

let widen_int (i1, r1, r1') (i2, r2, r2') =
  (widen_itv i1 i2, widen_rem d1 r1 r2, widen_rem d2 r1' r2')

let plus_int (i1, r1, r1') (i2, r2, r2') =
  (plus_itv i1 i2, plus_rem d1 r1 r2, plus_rem d2 r1' r2')

let mult_int (i1, r1, r1') (i2, r2, r2') =
  (mult_itv i1 i2, mult_rem d1 r1 r2, mult_rem d2 r1' r2')

let neg_int (i, r, r') = (neg_itv i, neg_rem d1 r, neg_rem d2 r')

let less_int (i1, r1, r1') (i2, r2, r2') =
  (less_itv i1 i2, less_rem r1 r2, less_rem r1' r2')

let not_int (i, r, r') = (not_itv i, not_rem r, not_rem r')

let and_int (i1, r1, r1') (i2, r2, r2') =
  (and_itv i1 i2, and_rem r1 r2, and_rem r1' r2')

let or_int (i1, r1, r1') (i2, r2, r2') =
  (or_itv i1 i2, or_rem r1 r2, or_rem r1' r2')

let maybe_t_int (i, r, r') = maybe_t_itv i && maybe_t_rem r && maybe_t_rem r'
let maybe_f_int (i, r, r') = maybe_f_itv i && maybe_f_rem r && maybe_t_rem r'

type abs_val = abs_int * LocSet.t

let join_loc l1 l2 = LocSet.union l1 l2
let join (n1, a1) (n2, a2) = (join_int n1 n2, join_loc a1 a2)
let widen (n1, a1) (n2, a2) = (widen_int n1 n2, join_loc a1 a2)
let plus (n1, _) (n2, _) = (plus_int n1 n2, LocSet.empty)
let mult (n1, _) (n2, _) = (mult_int n1 n2, LocSet.empty)
let neg (n, _) = (neg_int n, LocSet.empty)
let less (n1, _) (n2, _) = (less_int n1 n2, LocSet.empty)
let lnot (n, _) = (not_int n, LocSet.empty)
let andalso (n1, _) (n2, _) = (and_int n1 n2, LocSet.empty)
let orelse (n1, _) (n2, _) = (or_int n1 n2, LocSet.empty)
let maybe_t (n, _) = maybe_t_int n
let maybe_f (n, _) = maybe_f_int n

let eqb_aval ((i1, r1, r1'), a1) ((i2, r2, r2'), a2) =
  i1 = i2 && RemSet.equal r1 r2 && RemSet.equal r1' r2' && LocSet.equal a1 a2

type abs_mem = abs_val LocMap.t
type abstract_semantics = abs_mem LblMap.t

let string_of_value v = match v with Value n -> string_of_int n | Addr a -> a

let string_of_mem mem =
  let init = "[" in
  let f k v (acc, first) =
    let semicolon = if first then "" else " ; " in
    (acc ^ semicolon ^ k ^ " |-> " ^ string_of_value v, false)
  in
  let ret, _ = LocMap.fold f mem (init, true) in
  ret ^ "]"

let string_of_memset ms =
  let init = "{" in
  let f v (acc, first) =
    let comma = if first then "\n " else ",\n " in
    (acc ^ comma ^ string_of_mem v, false)
  in
  let ret, first = MemSet.fold f ms (init, true) in
  if first then ret ^ "}" else ret ^ "\n}"

let string_of_col col =
  let init = "" in
  let f k v acc = acc ^ string_of_int k ^ ":\n" ^ string_of_memset v ^ "\n" in
  LblMap.fold f col init

(* if we try to print all numbers from 0 to 231009, the OCaml runtime hangs up... *)
let string_of_rems ((rs1 : RemSet.t), (rs2 : RemSet.t)) =
  match (RemSet.mem r1 rs1, RemSet.mem r2 rs2) with
  | true, true -> (
    match (RemSet.cardinal rs1 = 1, RemSet.cardinal rs2 = 1) with
    | true, true -> "Watermarked"
    | _ -> "May be watermarked")
  | _ -> "Not watermarked"
(*
  let init = "{" in
  let f (acc, first) v =
    let comma = if first then "" else ", " in
    (acc ^ comma ^ string_of_int v, false)
  in
  let ret, _ = List.fold_left f (init, true) (RemSet.elements r) in
  ret ^ "}"
  *)

let string_of_aint (i, r1, r2) = string_of_itv i ^ ", " ^ string_of_rems (r1, r2)

let string_of_aloc (a : LocSet.t) =
  let init = "{" in
  let f v (acc, first) =
    let comma = if first then "" else ", " in
    (acc ^ comma ^ v, false)
  in
  let ret, _ = LocSet.fold f a (init, true) in
  ret ^ "}"

let string_of_aval (n, a) =
  "(" ^ string_of_aint n ^ ", " ^ string_of_aloc a ^ ")"

let string_of_amem amem =
  let init = "[" in
  let f k v (acc, first) =
    let semicolon = if first then "" else " ; " in
    (acc ^ semicolon ^ k ^ " |-> " ^ string_of_aval v, false)
  in
  let ret, _ = LocMap.fold f amem (init, true) in
  ret ^ "]"

let string_of_asem (asem : abstract_semantics) =
  let init = "" in
  let f k v acc = acc ^ string_of_int k ^ ":\n" ^ string_of_amem v ^ "\n" in
  LblMap.fold f asem init

let locmap_find_opt x m =
  match LocMap.find x m with exception Not_found -> None | v -> Some v

let rec interp_exp e (m : memory) =
  match e with
  | NUM n -> Some (Value n)
  | VAR x -> locmap_find_opt x m
  | DEREF x -> (
    match locmap_find_opt x m with
    | Some (Addr a) -> locmap_find_opt a m
    | _ -> None)
  | AT x -> Some (Addr x)
  | ADD (e1, e2) -> (
    match (interp_exp e1 m, interp_exp e2 m) with
    | Some (Value n1), Some (Value n2) -> Some (Value (n1 + n2))
    | _, _ -> None)
  | MULT (e1, e2) -> (
    match (interp_exp e1 m, interp_exp e2 m) with
    | Some (Value n1), Some (Value n2) -> Some (Value (n1 * n2))
    | _, _ -> None)
  | MINUS e -> (
    match interp_exp e m with Some (Value n) -> Some (Value (-n)) | _ -> None)
  | TRUE -> Some (Value 1)
  | FALSE -> Some (Value 0)
  | READ -> Some (Value (read_int ()))
  | NOT e -> (
    match interp_exp e m with
    | Some (Value n) -> if n = 0 then Some (Value 1) else Some (Value 0)
    | _ -> None)
  | ANDALSO (e1, e2) -> (
    match (interp_exp e1 m, interp_exp e2 m) with
    | Some (Value n1), Some (Value n2) ->
      if n1 * n2 = 0 then Some (Value 0) else Some (Value 1)
    | _, _ -> None)
  | ORELSE (e1, e2) -> (
    match (interp_exp e1 m, interp_exp e2 m) with
    | Some (Value n1), Some (Value n2) ->
      if n1 = 0 && n2 = 0 then Some (Value 0) else Some (Value 1)
    | _, _ -> None)
  | LESS (e1, e2) -> (
    match (interp_exp e1 m, interp_exp e2 m) with
    | Some (Value n1), Some (Value n2) ->
      if n1 < n2 then Some (Value 1) else Some (Value 0)
    | _, _ -> None)

let rec ainterp_exp e (m : abs_mem) : abs_val =
  match e with
  | NUM n ->
    let rem d = RemSet.singleton (rem n d) in
    ((Itv (Fin n, Fin n), rem d1, rem d2), LocSet.empty)
  | VAR x -> (
    match locmap_find_opt x m with
    | None -> ((Bot, RemSet.empty, RemSet.empty), LocSet.empty)
    | Some v -> v)
  | DEREF x -> (
    match locmap_find_opt x m with
    | None -> ((Bot, RemSet.empty, RemSet.empty), LocSet.empty)
    | Some (_, addrs) ->
      LocSet.fold
        (fun x acc ->
          let v =
            match locmap_find_opt x m with
            | None -> ((Bot, RemSet.empty, RemSet.empty), LocSet.empty)
            | Some v -> v
          in
          join v acc)
        addrs
        ((Bot, RemSet.empty, RemSet.empty), LocSet.empty))
  | AT x -> ((Bot, RemSet.empty, RemSet.empty), LocSet.singleton x)
  | ADD (e1, e2) -> plus (ainterp_exp e1 m) (ainterp_exp e2 m)
  | MULT (e1, e2) -> mult (ainterp_exp e1 m) (ainterp_exp e2 m)
  | MINUS e -> neg (ainterp_exp e m)
  | TRUE -> ((itv_true, rem_true, rem_true), LocSet.empty)
  | FALSE -> ((itv_false, rem_false, rem_true), LocSet.empty)
  | READ -> (top_int, LocSet.empty)
  | NOT e -> lnot (ainterp_exp e m)
  | ANDALSO (e1, e2) -> andalso (ainterp_exp e1 m) (ainterp_exp e2 m)
  | ORELSE (e1, e2) -> orelse (ainterp_exp e1 m) (ainterp_exp e2 m)
  | LESS (e1, e2) -> less (ainterp_exp e1 m) (ainterp_exp e2 m)

(* d execution *)
type cfg = int -> int

let lbl_to_cmd : (int, cmd) Hashtbl.t = Hashtbl.create 10
let bind (f : cfg) (lbl : int) (c : int) l = if l = lbl then c else f l
let label_of (LBL (l, _)) = l

let rec make_cfg (c : program) (l_exit : int) next nextTrue nextFalse :
    cfg * cfg * cfg =
  match c with
  | LBL (l, c) -> (
    Hashtbl.add lbl_to_cmd l c;
    match c with
    | SEQ (c1, c2) ->
      let next = bind next l (label_of c1) in
      let next, nextTrue, nextFalse =
        make_cfg c1 (label_of c2) next nextTrue nextFalse
      in
      make_cfg c2 l_exit next nextTrue nextFalse
    | IF (_, c1, c2) ->
      let nextTrue = bind nextTrue l (label_of c1) in
      let nextFalse = bind nextFalse l (label_of c2) in
      let next, nextTrue, nextFalse =
        make_cfg c1 l_exit next nextTrue nextFalse
      in
      make_cfg c2 l_exit next nextTrue nextFalse
    | WHILE (_, c) ->
      (* while (not e) c *)
      let nextTrue = bind nextTrue l l_exit in
      let nextFalse = bind nextFalse l (label_of c) in
      make_cfg c l next nextTrue nextFalse
    | GOTO _ -> (next, nextTrue, nextFalse)
    | ASSIGN _ ->
      let next = bind next l l_exit in
      (next, nextTrue, nextFalse)
    | PTRASSIGN _ ->
      let next = bind next l l_exit in
      (next, nextTrue, nextFalse))

let rec dintp_aux l m next nextTrue nextFalse =
  match Hashtbl.find lbl_to_cmd l with
  | exception Not_found -> m
  | c -> (
    match c with
    | IF (e, _, _) | WHILE (e, _) -> (
      match interp_exp e m with
      | Some (Value n) ->
        if n = 0 then dintp_aux (nextFalse l) m next nextTrue nextFalse
        else dintp_aux (nextTrue l) m next nextTrue nextFalse
      | _ -> m)
    | SEQ (_, _) -> dintp_aux (next l) m next nextTrue nextFalse
    | GOTO e -> (
      match interp_exp e m with
      | Some (Value l) -> dintp_aux l m next nextTrue nextFalse
      | _ -> m)
    | ASSIGN (x, e) -> (
      match interp_exp e m with
      | Some v -> dintp_aux (next l) (LocMap.add x v m) next nextTrue nextFalse
      | _ -> m)
    | PTRASSIGN (x, e) -> (
      match (locmap_find_opt x m, interp_exp e m) with
      | Some (Addr a), Some v ->
        dintp_aux (next l) (LocMap.add a v m) next nextTrue nextFalse
      | _ -> m))

let dintp pgm =
  let init _ = -1 in
  Hashtbl.clear lbl_to_cmd;
  let next, nextTrue, nextFalse = make_cfg pgm (-1) init init init in
  let init_mem = LocMap.empty in
  dintp_aux (label_of pgm) init_mem next nextTrue nextFalse

let rec dcintp_aux l m acc next nextTrue nextFalse =
  let acc =
    let single_map = LblMap.add l (MemSet.singleton m) LblMap.empty in
    LblMap.merge
      (fun _ v1 v2 ->
        match (v1, v2) with
        | None, None -> None
        | Some v, None -> Some v
        | None, Some v -> Some v
        | Some v1, Some v2 -> Some (MemSet.union v1 v2))
      acc single_map
  in
  match Hashtbl.find lbl_to_cmd l with
  | exception Not_found -> acc
  | c -> (
    match c with
    | IF (e, _, _) | WHILE (e, _) -> (
      match interp_exp e m with
      | Some (Value n) ->
        if n = 0 then dcintp_aux (nextFalse l) m acc next nextTrue nextFalse
        else dcintp_aux (nextTrue l) m acc next nextTrue nextFalse
      | _ -> acc)
    | SEQ (_, _) -> dcintp_aux (next l) m acc next nextTrue nextFalse
    | GOTO e -> (
      match interp_exp e m with
      | Some (Value l) -> dcintp_aux l m acc next nextTrue nextFalse
      | _ -> acc)
    | ASSIGN (x, e) -> (
      match interp_exp e m with
      | Some v ->
        dcintp_aux (next l) (LocMap.add x v m) acc next nextTrue nextFalse
      | _ -> acc)
    | PTRASSIGN (x, e) -> (
      match (locmap_find_opt x m, interp_exp e m) with
      | Some (Addr a), Some v ->
        dcintp_aux (next l) (LocMap.add a v m) acc next nextTrue nextFalse
      | _ -> acc))

(* d collecting semantics *)
let dcintp pgm : collecting_semantics =
  let init _ = -1 in
  Hashtbl.clear lbl_to_cmd;
  let next, nextTrue, nextFalse = make_cfg pgm (-1) init init init in
  let init_mem = LocMap.empty in
  dcintp_aux (label_of pgm) init_mem LblMap.empty next nextTrue nextFalse

let merge_mem merge (m1 : abs_mem) (m2 : abs_mem) : abs_mem =
  let m _ v1 v2 =
    match (v1, v2) with
    | None, None -> None
    | Some v1, None -> Some v1
    | None, Some v2 -> Some v2
    | Some v1, Some v2 -> Some (merge v1 v2)
  in
  LocMap.merge m m1 m2

(* strong update *)
let strong_update (m : abs_mem) (x : loc) (v : abs_val) : abs_mem =
  LocMap.add x v m

let join_mem (m1 : abs_mem) (m2 : abs_mem) : abs_mem = merge_mem join m1 m2
let widen_mem (m1 : abs_mem) (m2 : abs_mem) : abs_mem = merge_mem widen m1 m2

let filter_rem_with_itv l r d x =
  let l, r = (rem l d, rem r d) in
  if l <= r then l <= x && x <= r else l <= x || x <= r

let meet_with_filter (m : abs_mem) (filter : Lia.filter) : abs_mem =
  let f acc (x, cstr) =
    match locmap_find_opt x acc with
    | None -> LocMap.empty
    | Some ((i, r1, r2), _) -> (
      match meet_itv i cstr with
      | Bot -> LocMap.empty
      | Itv (Fin l, Fin r) as i ->
        let r1 =
          if r - l + 1 >= d1 then r1
          else RemSet.filter (filter_rem_with_itv l r d1) r1
        in
        let r2 =
          if r - l + 1 >= d2 then r2
          else RemSet.filter (filter_rem_with_itv l r d2) r2
        in
        LocMap.add x ((i, r1, r2), LocSet.empty) acc
      | i -> LocMap.add x ((i, r1, r2), LocSet.empty) acc)
  in
  List.fold_left f m filter

let filter_true (e : exp) (m : abs_mem) =
  let true_filter, _ = Lia.make_filter e in
  meet_with_filter m true_filter

let filter_false (e : exp) (m : abs_mem) =
  let _, false_filter = Lia.make_filter e in
  meet_with_filter m false_filter

let goto_targets = ref RemSet.empty
let update_goto_targets t = goto_targets := RemSet.add t !goto_targets

let ainterp_cmd (l, m) (sem : abstract_semantics) next nextTrue nextFalse :
    abstract_semantics * (int * abs_mem) list =
  let m_orig, new_entry =
    match LblMap.find l sem with
    | exception Not_found -> (LocMap.empty, true)
    | m_orig -> (m_orig, false)
  in
  let widen =
    match Hashtbl.find lbl_to_cmd l with
    | exception Not_found -> false
    | WHILE _ -> true
    | _ -> RemSet.mem l !goto_targets
  in
  let m = if widen then widen_mem m_orig m else join_mem m_orig m in
  let sem = LblMap.add l m sem in
  let not_changed = LocMap.equal eqb_aval m_orig m && not new_entry in
  if not_changed then (sem, [])
  else
    let worklist =
      match Hashtbl.find lbl_to_cmd l with
      | exception Not_found -> []
      | IF (e, _, _) | WHILE (e, _) ->
        let condition = ainterp_exp e m in
        let if_true =
          if maybe_t condition then filter_true e m else LocMap.empty
        in
        let if_false =
          if maybe_f condition then filter_false e m else LocMap.empty
        in
        [(nextTrue l, if_true); (nextFalse l, if_false)]
      | SEQ (_, _) -> [(next l, m)]
      | GOTO e ->
        let (i, r1, r2), _ = ainterp_exp e m in
        let test_target t =
          let test_itv =
            match i with
            | Itv (NegInf, Inf) -> true
            | Itv (NegInf, Fin r) -> t <= r
            | Itv (Fin l, Inf) -> l <= t
            | Itv (Fin l, Fin r) -> l <= t && t <= r
            | _ -> false
          in
          let test_rem1 = RemSet.mem (rem t d1) r1 in
          let test_rem2 = RemSet.mem (rem t d2) r2 in
          test_itv && test_rem1 && test_rem2
        in
        let targets =
          Hashtbl.fold
            (fun t _ acc ->
              if test_target t then
                let () = update_goto_targets t in
                t :: acc
              else acc)
            lbl_to_cmd []
        in
        let may_exit =
          match i with
          | Itv (Fin l, Fin r) ->
            (* assume that labels are contiguous *)
            not (Hashtbl.mem lbl_to_cmd l && Hashtbl.mem lbl_to_cmd r)
          | _ -> true
        in
        let targets = if may_exit then -1 :: targets else targets in
        List.map (fun l -> (l, m)) targets
      | ASSIGN (x, e) ->
        let v = ainterp_exp e m in
        let m = strong_update m x v in
        [(next l, m)]
      | PTRASSIGN (x, e) ->
        let m =
          match locmap_find_opt x m with
          | None -> LocMap.empty
          | Some (_, addrs) ->
            let v = ainterp_exp e m in
            LocSet.fold
              (fun a acc -> join_mem (strong_update m a v) acc)
              addrs LocMap.empty
        in
        [(next l, m)]
    in
    (sem, worklist)

(* d analysis *)
(* -1 = exit label *)
let analysis pgm : abstract_semantics =
  let init _ = -1 in
  Hashtbl.clear lbl_to_cmd;
  let next, nextTrue, nextFalse = make_cfg pgm (-1) init init init in
  let rec loop sem worklist =
    match worklist with
    | [] -> sem
    | hd :: tl ->
      let sem, worklist = ainterp_cmd hd sem next nextTrue nextFalse in
      loop sem (List.rev_append worklist tl)
  in
  loop LblMap.empty [(label_of pgm, LocMap.empty)]
